# Nvidia VirtIO Net Controller Daemon

VirtIO net controller enables users to create VirtIO-net emulated PCIe devices
in the system where the NVIDIA® BlueField® DPU is connected. This is done by the
virtio-net-controller software module present in the DPU. Virtio-net emulated
devices allow users to hot plug virtio-net PCIe PF Ethernet NIC devices and
virtio-net PCI VF Ethernet NIC devices in the host system where the DPU is
plugged in. By using this solution, virtio net device offloads traffic handling
from host CPU to the NIC of DPU.

# System Preparation

Use BlueField-2 as example, mst device should be adjusted accordingly otherwise.

## Prepare DPU for hotplug devices:
```sh
$ mst start
$ mlxconfig -d /dev/mst/mt41686_pciconf0 s \
		PF_BAR2_ENABLE=0 \
		PER_PF_NUM_SF=1
$ mlxconfig -d /dev/mst/mt41686_pciconf0 s \
		PCI_SWITCH_EMULATION_ENABLE=1 \
		PCI_SWITCH_EMULATION_NUM_PORT=16 \
		VIRTIO_NET_EMULATION_ENABLE=1 \
		VIRTIO_NET_EMULATION_NUM_VF=0 \
		VIRTIO_NET_EMULATION_NUM_PF=0 \
		VIRTIO_NET_EMULATION_NUM_MSIX=16 \
		VIRTIO_NET_EMULATION_VENDOR_ID=0x1af4 \
		VIRTIO_NET_EMULATION_DEVICE_ID=0x1041 \
		VIRTIO_NET_EMULATION_CLASS_CODE=0x028000 \
		ECPF_ESWITCH_MANAGER=1 \
		ECPF_PAGE_SUPPLIER=1 \
		SRIOV_EN=0 \
		PF_SF_BAR_SIZE=8 \
		PF_TOTAL_SF=64
$ mlxconfig -d /dev/mst/mt41686_pciconf0.1 s \
		PF_SF_BAR_SIZE=10 \
		PF_TOTAL_SF=64
```
Cold reboot host

## Prepare DPU for static devices with SR-IOV support:
```sh
$ mst start
$ mlxconfig -d /dev/mst/mt41686_pciconf0 s \
		PF_BAR2_ENABLE=0 \
		PER_PF_NUM_SF=1
$ mlxconfig -d /dev/mst/mt41686_pciconf0 s \
		PCI_SWITCH_EMULATION_ENABLE=1 \
		PCI_SWITCH_EMULATION_NUM_PORT=16 \
		VIRTIO_NET_EMULATION_ENABLE=1 \
		VIRTIO_NET_EMULATION_NUM_VF=126 \
		VIRTIO_NET_EMULATION_NUM_PF=4 \
		VIRTIO_NET_EMULATION_NUM_MSIX=4 \
		VIRTIO_NET_EMULATION_VENDOR_ID=0x1af4 \
		VIRTIO_NET_EMULATION_DEVICE_ID=0x1041 \
		VIRTIO_NET_EMULATION_CLASS_CODE=0x028000 \
		ECPF_ESWITCH_MANAGER=1 \
		ECPF_PAGE_SUPPLIER=1 \
		SRIOV_EN=1 \
		PF_SF_BAR_SIZE=8 \
		PF_TOTAL_SF=508
$ mlxconfig -d /dev/mst/mt41686_pciconf0.1 s \
		PF_TOTAL_SF=1 \
		PF_SF_BAR_SIZE=8
```
Cold reboot host

**NOTE**:
1. To use static device, it is recommended to blacklist kernel module
virtio_pci and virtio_net from guest OS before proceeding. If those modules
are built into kernel image, they are not blacklistable. In such case, it's
recommended for users to have a way to start the virtio_net_controller from
DPU using SSH or serial/oob interfaces. This can be achieved by connecting
the RShim/serial cable to another host, or setting up the OOB interface.

    Normally, this is not needed as controller daemon starts automatically when
system boots.

2. For the systems which BIOS is not doing address space allocation for
unpopulated downstream ports of the pcie switch, __"pci=realloc"__ is needed
inside boot commandline.

    It is recommended to add __"pci=assign-busses"__ to boot command line when
creating more than 127 VFs. Without this option, errors below might appear
from host. And virio driver won't probe those devices as type is incorrect.
    ```sh
    [  617.382854] pci 0000:84:00.0: [1af4:1041] type 7f class 0xffffff
    [  617.382883] pci 0000:84:00.0: unknown header type 7f, ignoring device
    ```

# Building

Dependency:
    mlnx-ofed, mlnx-libsnap, libev, libev-devel, libmnl-devel

To build and install this daemon, depends on the packet, there are different
steps.

## Release Tarball

Follow steps below if you received the release tarball after untar:
```sh
$ ./configure --prefix=/usr && make -j 8 && make install
```

## Git Repository

Follow steps below if you're developing on top of the git repository.
```sh
$ ./autogen.sh && ./configure --prefix=/usr && make && make install
```
Typically the autogen and configure steps only need to be done the first time
unless configure.ac or Makefile.am changes.

To build rpm/deb:
```sh
$ scripts/build_rpm.sh
$ tar -xvf *.gz
$ cd virtio-net-controller-<ver>
$ dpkg-buildpackage
```

# Usage

The controller has a systemd service running and a user interface tool to
communicate with the service.

* Service: virtio-net-controller.service
* User Interface: virtnet
* Log: controller is using system log system. Run:
```sh
  $ journalctl -u virtio-net-controller
```

## Controller Service

If controller is running on a DPU, the service is enabled by default. It starts
automatically. Run command below to check the status.

```sh
$ systemctl status virtio-net-controller.service
```

If daemon is not running, start controller by running command below.
Make sure to check the status after command starts.

```sh
$ systemctl start virtio-net-controller.service
```

To stop the daemon, first unload virtio-net/virtio-pci from host. Then run
command below from DPU:

```sh
$ systemctl stop virtio-net-controller.service
```

To enable the daemon automatically start, run:

```sh
$ systemctl enable virtio-net-controller.service
```

## Controller Configuration File

A config file __/opt/mellanox/mlnx_virtnet/virtnet.conf__ can be used to set
parameters for the service. If config file is changed while controller service
is running, the new config won't take effect until service is reloaded.

Currently supported options:

* ib_dev_for_static_pf: the RDMA device(e.g, mlx5_0) which the static virtIO PF will be created on
* is_lag: whether or not LAG is used. Note that if lag is used, make sure to use the correct IB dev for static PF
* pf_mac: the base MAC address for static PFs. MACs are automatically assigned with following pattern: pf_mac->pf_0, pf_mac + 1 -> pf_1, etc. Note that controller won't validate the MAC address (other than the length), user must ensure MAC is valid and unique

Here is an example (JSON format):

```json
{
  "ib_dev_for_static_pf": "mlx5_0",
  "is_lag": 0,
  "pf_mac": "00:11:22:33:44:55"
}
```
    
## Controller User Interface

Each command has its own help manual, e.g, _virtnet list -h_

1. list current devices

```sh
$ virtnet list
```

2. Hotplug virtio net device with 1500 mtu and 3 virtio queues of depth 1024 entries with vlan control support, data path relies on mlx5_0:

```sh
$ virtnet hotplug -i mlx5_0 -f 0x80000 -m 0C:C4:7A:FF:22:93 -t 1500 -n 3 -s 1024
```
Max number of virtio queues are bounded by the minimum of the numbers below

* `VIRTIO_NET_EMULATION_NUM_MSIX` from `mlxconfig -d <mst_dev> q`
* `max_virtq` from `virtnet list`

If hotplug succeeds, hotplug device info returns similar as below:

```json
{
  "bdf": "84:0.0",
  "id": 1,
  "sf_rep_net_device": "en3f0pf0sf1001",
  "mac": "0C:C4:7A:FF:22:93"
}
```
Add the SF interface to corresponding OVS bridge:

```sh
$ ovs add-port <bridge> en3f0pf0sf1001
```

3. Modify the MAC address of virtio physical device ID 0

```sh
$ virtnet modify -p 0 device -m 0C:C4:7A:FF:22:93
```

4. Modify the MSI-X number of VF device

```sh
virtnet modify -p 0 -v 0 device -n 8
```

5. Modify the queue options of virtio physical device ID 0

```sh
$ virtnet modify -p 0 queue -e event -n 10 -c 30
```

6. Unplug virtio physical device ID 0, only applicable to hotplug device

```sh
$ virtnet unplug -p 0
```

## Limitations
1. When port MTU (p0/p1 from DPU) is changed(e.g, from 1500 to 9000) after
controller starts, it is required to restart controller service.
